from genpeds.downloader import scrape_ipeds_data
import argparse

'''simple and doesn't handle many errors at the moment,
but it'll do for now.'''

def parse_years(year_str):
    '''parse year argument
    
    years take:
    1. range, ie. 2001-2023
    2. list, ie. [2022,2023,1990]
    3. single int, ie. 2022
    '''
    if len(year_str) == 1 and '-' not in year_str[0]:
        yrs_iter = int(year_str[0])
    elif len(year_str) == 1 and '-' in year_str[0]:
        split_yrs = year_str[0].split('-')
        start, end = split_yrs
        yrs_iter = (
            int(start), int(end)
        )
    elif len(year_str) > 1:
        yrs_iter = [int(yr) for yr in year_str]
    
    return yrs_iter

def main():
    parser = argparse.ArgumentParser('genpeds', description='NCES IPEDS subject-data scraper')
    parser.add_argument('subject', 
                        choices=['characteristics', 'admissions', 'enrollment', 
                                            'cip', 'completion', 'graduation'],
                        help=('IPEDS data subject to download. Options include:\n'
                        '[characteristics, admissions, enrollment, cip, completion, graduation]'))
    parser.add_argument('-y', '--years',
                        required=True, 
                        nargs='+',
                        help=('years of IPEDS subject data to download. Can take '
                              'range, ie. 2001-2023 or list, ie. 2022 2023 1990 or single year, ie. 2017'))
    
    args = parser.parse_args()
    cleaned_yrs = parse_years(args.years) 
    scrape_ipeds_data(
        subject=args.subject,
        year_range=cleaned_yrs
    )
    
if __name__ == '__main__':
    main()


from genpeds.downloader import scrape_ipeds_data
from genpeds.cleaners import CLEANERS
from genpeds.config import DATASETS, VARIABLE_DICT
import pandas as pd

import shutil
from abc import ABC, abstractmethod

class IPDS(ABC):
    subject = None
    
    def __init__(self, year_range=None):
        self.year_range = year_range # year range by user
        self.available_years = DATASETS[self.subject]['years_available']
        self.variable_dict = VARIABLE_DICT[self.subject]

    def get_description(self):
        '''returns description of the subject data.'''
        return DATASETS[self.subject]['description']
    
    def get_available_years(self):
        '''returns available years for a subject's data.'''
        return self.available_years
    
    def get_available_vars(self):
        '''returns dict of available variables for a subject and their descriptions.'''
        return self.variable_dict
    
    def lookup_var(self, var=''):
        '''returns variable description.'''
        return self.variable_dict[var]

    def scrape(self, see_progress=False):
        '''downloads NCES IPEDS data to disk on specified years for a defined subject.
        
        :param see_progress::
            (bool) prints completion statement for extraction of each year's data. If False, no messages printed.
        '''
        scrape_ipeds_data(subject=self.subject, year_range=self.year_range, see_progress=see_progress)

    @abstractmethod
    def clean(self):
        '''clean the data'''
        pass

    @abstractmethod
    def run(self):
        '''scrape and clean'''
        pass


class Characteristics(IPDS):
    '''IPEDS Characteristics'''
    subject = 'characteristics'

    def __init__(self, year_range=None):
        '''IPEDS Characteristics data.
        
        :param year_range::
          tuple of inclusive year integers (indicates a range), iterable of year integers (indicates group of individual years), or single year to pull data from.
        
          ex. year_range=(2002,2012)

        -----------------  
        <h3>Example Use:</h3>
        >>> import genpeds as ed
        >>> chars_2000 = ed.Characteristics(year_range=[2000,2005,2010]) # three years of data
        >>> chars_2000.get_available_years()
         (1984,2023) # available years for Characteristics data
        >>> chars_data = chars_2000.run() # returns Pandas dataframe

        
        ----------------
        <h3>IPEDS</h3>
        The Integrated Postsecondary Education Data System (IPEDS), ran by the National Center for Education Statistics (NCES), is a collection of surveys annually conducted. All postsecondary institutions that participate in federal student aid financial aid programs are required to participate in these surveys. IPEDS covers eight subjects: 
        1. institutional characteristics
        2. admissions
        3. enrollment
        4. degrees and certificates conferred
        5. student persistence and success
        6. institutional prices
        7. student financial aid
        8. institutional resources including human, resources, finance, and academic libraries

        As of this version, `genpeds` only provides objects for the first five subject areas, as these areas provide data by gender and variables of interest like graduation rates and enrollment.
        '''
        super().__init__(year_range)

    def clean(self, char_dir='characteristicsdata', rm_disk=False) -> pd.DataFrame:
        '''cleans downloaded Characteristics data, returns Pandas Dataframe.
        
        :param char_dir::
          directory where raw Charactetistics data is located; defaults to default download dir name.
        :param rm_disk::
          removes downloaded Characteristics data from disk, after cleaning.
        '''
        df = CLEANERS[self.subject](char_dir)
        if rm_disk:
            shutil.rmtree(char_dir)
        return df
    
    def run(self, see_progress=False, rm_disk=False) -> pd.DataFrame:
        '''scrapes and cleans IPEDS Characteristics data; returns Pandas Dataframe.
        
        :param see_progress::
        (bool) When True, prints successful download confirmation for each year's data. If False, no messages printed.
        
        :param rm_disk::
        removes downloaded Admissions data from disk after data is cleaned and returned.
        '''
        self.scrape(see_progress=see_progress)
        df = self.clean(rm_disk=rm_disk)
        return df


class Admissions(IPDS):
    '''IPEDS Admissions'''
    subject = 'admissions'

    def __init__(self, year_range=None):
        '''IPEDS Admissions data.
        
        :param year_range::
          tuple of inclusive year integers (indicates a range), iterable of year integers (indicates group of individual years), or single year to pull data from.
          
          ex. year_range=(2002,2012)
          
        -----------------  
        <h3>Example Use:</h3>
        >>> import genpeds as ed
        >>> adm_2010s = ed.Admissions(year_range=(2010,2019)) # ten years of data
        >>> adm_2010s.get_available_years()
         (2001,2023) # available years for Admissions data
        >>> adm_data = adm_2010s.run() # returns Pandas dataframe

        ----------------
        <h3>IPEDS</h3>
        The Integrated Postsecondary Education Data System (IPEDS), ran by the National Center for Education Statistics (NCES), is a collection of surveys annually conducted. All postsecondary institutions that participate in federal student aid financial aid programs are required to participate in these surveys. IPEDS covers eight subjects: 
        1. institutional characteristics
        2. admissions
        3. enrollment
        4. degrees and certificates conferred
        5. student persistence and success
        6. institutional prices
        7. student financial aid
        8. institutional resources including human, resources, finance, and academic libraries

        As of this version, `genpeds` only provides objects for the first five subject areas, as these areas provide data by gender and variables of interest like graduation rates and enrollment.
        '''
        super().__init__(year_range)

    def clean(self, admit_dir='admissionsdata', rm_disk=False) -> pd.DataFrame:
        '''cleans downloaded Admissions data, returns Pandas Dataframe.
        
        :param admit_dir::
          directory where raw Admissions data is located; defaults to default download dir name.
        :param rm_disk::
          removes downloaded Admissions data from disk, after cleaning.
        '''
        df = CLEANERS[self.subject](admissions_dir=admit_dir)
        if rm_disk:
            shutil.rmtree(admit_dir) # removes data from disk
        return df
    
    def run(self, see_progress=False, merge_with_char=False, rm_disk=False) -> pd.DataFrame:
        '''scrapes and cleans Admissions data; returns Pandas Dataframe.
        
        :param see_progress::
        (bool) When True, prints successful download confirmation for each year's data. If False, no messages printed.

        :param merge_with_char::
        (bool) When True, scrapes Admissions data and merges with Characteristics data (includes variables like school name and address). Characteristics data will automatically be removed from disk when the Dataframe is returned. (To keep the Characteristics data, create a Characteristics object.)
        
        :param rm_disk::
        removes downloaded Admissions data from disk after data is cleaned and returned.
        '''
        self.scrape(see_progress=see_progress)
        df = self.clean(rm_disk=rm_disk)
        if merge_with_char:
            char_df = Characteristics(year_range=self.year_range).run(see_progress=see_progress, rm_disk=True)
            df = df.merge(char_df, on=['id', 'year'])
        return df
    

class Enrollment(IPDS):
    '''IPEDS Enrollment'''
    subject = 'enrollment'

    def __init__(self, year_range=None):
        '''IPEDS Enrollment data.
        
        :param year_range::
          tuple of inclusive year integers (indicates a range), iterable of year integers (indicates group of individual years), or single year to pull data from.
        
         ex. year_range=(2002,2012)

        -----------------  
        <h3>Example Use:</h3>
        >>> import genpeds as ed
        >>> enroll_2022 = ed.Enrollment(year_range=2022) # one year of data
        >>> enroll_2022.get_available_years()
         (1984,2023) # available years for Enrollment data
        >>> enroll_data = enroll_2022.run() # returns Pandas dataframe

        ----------------
        <h3>IPEDS</h3>
        The Integrated Postsecondary Education Data System (IPEDS), ran by the National Center for Education Statistics (NCES), is a collection of surveys annually conducted. All postsecondary institutions that participate in federal student aid financial aid programs are required to participate in these surveys. IPEDS covers eight subjects: 
        1. institutional characteristics
        2. admissions
        3. enrollment
        4. degrees and certificates conferred
        5. student persistence and success
        6. institutional prices
        7. student financial aid
        8. institutional resources including human, resources, finance, and academic libraries

        As of this version, `genpeds` only provides objects for the first five subject areas, as these areas provide data by gender and variables of interest like graduation rates and enrollment.
        '''
        super().__init__(year_range)

    def clean(self, student_level='undergrad', enroll_dir='enrollmentdata', rm_disk=False) -> pd.DataFrame:
        '''cleans downloaded Fall Enrollment data, returns Pandas Dataframe.
        
        :param student_level::
         level of student enrollment; options include ['undergrad', 'grad'].
        :param enroll_dir::
          directory where raw Enrollment data is located; defaults to default download dir name.
        :param rm_disk::
          removes downloaded Enrollment data from disk, after cleaning.
        '''
        df = CLEANERS[self.subject](enrollment_dir=enroll_dir, student_level=student_level)
        if rm_disk:
            shutil.rmtree(enroll_dir)
        return df
    
    def run(self, student_level='undergrad', see_progress=False, merge_with_char=False, rm_disk=False) -> pd.DataFrame:
        '''scrapes and cleans IPEDS Fall Enrollment data; returns Pandas Dataframe.
        
        :param student_level::
         level of student enrollment; options include ['undergrad', 'grad'].

        :param see_progress::
        (bool) When True, prints successful download confirmation for each year's data. If False, no messages printed.

        :param merge_with_char::
        (bool) When True, scrapes Enrollment data and merges with Characteristics data (includes variables like school name and address). Characteristics data will automatically be removed from disk when the Dataframe is returned. (To keep the Characteristics data, create a Characteristics object.)
        
        :param rm_disk::
        removes downloaded Enrollment data from disk after data is cleaned and returned.
        '''
        self.scrape(see_progress=see_progress)
        df = self.clean(rm_disk=rm_disk, student_level=student_level)
        if merge_with_char:
            char_df = Characteristics(year_range=self.year_range).run(see_progress=see_progress, rm_disk=True)
            df = df.merge(char_df, on=['id', 'year'])
        return df


class Cip(IPDS):
    '''CIP Codes'''
    subject = 'cip'

    def __init__(self, year_range=(1984,2023)):
        '''IPEDS CIP Codes data.

        :param year_range::
          tuple of inclusive year integers (indicates a range), iterable of year integers (indicates group of individual years), or single year to pull data from.

        CIP, or Classification of Instructional Programs, are key-value pairs for subject study fields. CIP's vary by year, and are relevant to identify subject field in completion data. Available for years 1984-2023. CIP data should be used in conjunction with Completion data.
        '''
        super().__init__(year_range)

    def clean(self, cip_dir='cipdata', rm_disk=False) -> pd.DataFrame:
        '''cleans downloaded CIP data, returns Pandas Dataframe.
        
        :param cip_dir::
          directory where raw CIP data is located; defaults to default download dir name.
        :param rm_disk::
          removes downloaded CIP data from disk, after cleaning.
        '''
        df = CLEANERS[self.subject](cip_codes_dir=cip_dir)
        if rm_disk:
            shutil.rmtree(cip_dir)
        return df
    
    def run(self, see_progress=False, rm_disk=False) -> pd.DataFrame:
        '''scrapes and cleans IPEDS CIP data; returns Pandas DataFrame.

        :param see_progress::
        (bool) When True, prints successful download confirmation for each year's data. If False, no messages printed.
        
        :param rm_disk::
        removes downloaded Enrollment data from disk after data is cleaned and returned.
        '''
        self.scrape(see_progress=see_progress)
        df = self.clean(rm_disk=rm_disk)
        return df


class Completion(IPDS):
    '''IPEDS Completion'''
    subject = 'completion'

    def __init__(self, year_range=(1984,2023)):
        '''IPEDS Completion data.
        
        :param year_range::
          tuple of inclusive year integers (indicates a range), iterable of year integers (indicates group of individual years), or single year to pull data from.
        
         ex. year_range=(2002,2012)
        
        -----------------  
        <h3>Example Use:</h3>
        >>> import genpeds as ed
        >>> complete_2022 = ed.Completion(year_range=2022) # one year of data
        >>> complete_2022.get_available_years()
         (1984,2023) # available years for Enrollment data
        >>> complete_data = complete_2022.run() # returns Pandas dataframe

        ----------------
        <h3>IPEDS</h3>
        The Integrated Postsecondary Education Data System (IPEDS), ran by the National Center for Education Statistics (NCES), is a collection of surveys annually conducted. All postsecondary institutions that participate in federal student aid financial aid programs are required to participate in these surveys. IPEDS covers eight subjects: 
        1. institutional characteristics
        2. admissions
        3. enrollment
        4. degrees and certificates conferred
        5. student persistence and success
        6. institutional prices
        7. student financial aid
        8. institutional resources including human, resources, finance, and academic libraries

        As of this version, `genpeds` only provides objects for the first five subject areas, as these areas provide data by gender and variables of interest like graduation rates and enrollment.
        '''
        super().__init__(year_range)

    def clean(self, degree_level='bach', complete_dir='completiondata', rm_disk=False) -> pd.DataFrame:
        '''cleans downloaded Completion data, returns Pandas Dataframe.
        
        :param degree_level::
         level of student degree completion; options include ['assc', 'bach', 'mast', 'doct'].
        :param complete_dir::
          directory where raw Completion data is located; defaults to default download dir name.
        :param rm_disk::
          removes downloaded Completion data from disk, after cleaning.
        '''
        df = CLEANERS[self.subject](completion_dir=complete_dir, level=degree_level)
        if rm_disk:
            shutil.rmtree(complete_dir)
        return df
    
    def run(self, degree_level='bach', see_progress=False, merge_with_char=False, get_cip_codes=True, rm_disk=False) -> pd.DataFrame:
        '''scrapes and cleans IPEDS Completion data; returns Pandas Dataframe.
        
        :param degree_level::
         level of student degree completion; options include ['assc', 'bach', 'mast', 'doct'].

        :param see_progress::
        (bool) When True, prints successful download confirmation for each year's data. If False, no messages printed.

        :param merge_with_char::
        (bool) When True, scrapes Completion data and merges with Characteristics data (includes variables like school name and address). Characteristics data will automatically be removed from disk when the Dataframe is returned. (To keep the Characteristics data, create a Characteristics object.)
        
        :param get_cip_codes::
        (bool) When True, scrapes CIP (e.g., field of study) codes/labels and merges with Completion data.

        :param rm_disk::
        removes downloaded Enrollment data from disk after data is cleaned and returned.
        '''
        self.scrape(see_progress=see_progress)
        df = self.clean(rm_disk=rm_disk, degree_level=degree_level)
        if merge_with_char:
            char_df = Characteristics(year_range=self.year_range).run(see_progress=see_progress, rm_disk=True)
            df = df.merge(char_df, on=['id', 'year'])
        if get_cip_codes:
            cip_df = Cip(year_range=self.year_range).run(see_progress=see_progress, rm_disk=True)
            df = df.merge(cip_df, on=['cip', 'year'])
        return df


class Graduation(IPDS):
    '''IPEDS Graduation'''
    subject = 'graduation'

    def __init__(self, year_range=(1984,2023)):
        '''IPEDS Graduation data.
        
        :param year_range::
          tuple of inclusive year integers (indicates a range), iterable of year integers (indicates group of individual years), or single year to pull data from.
        
         ex. year_range=(2002,2012)
        
        -----------------  
        <h3>Example Use:</h3>
        >>> import genpeds as ed
        >>> grad_aughts = ed.Completion(year_range=(2000,2009)) # ten years of data
        >>> grad_aughts.get_available_years()
         (2000,2023) # available years for Enrollment data
        >>> grad_data = grad_aughts.run() # returns Pandas dataframe

        ----------------
        <h3>IPEDS</h3>
        The Integrated Postsecondary Education Data System (IPEDS), ran by the National Center for Education Statistics (NCES), is a collection of surveys annually conducted. All postsecondary institutions that participate in federal student aid financial aid programs are required to participate in these surveys. IPEDS covers eight subjects: 
        1. institutional characteristics
        2. admissions
        3. enrollment
        4. degrees and certificates conferred
        5. student persistence and success
        6. institutional prices
        7. student financial aid
        8. institutional resources including human, resources, finance, and academic libraries

        As of this version, `genpeds` only provides objects for the first five subject areas, as these areas provide data by gender and variables of interest like graduation rates and enrollment.
        '''
        super().__init__(year_range)

    def clean(self, degree_level='bach', grad_dir='graduationdata', rm_disk=False) -> pd.DataFrame:
        '''cleans downloaded undergraduate Graduation data, returns Pandas Dataframe.
        
        :param degree_level::
         level of graduate; options include ['assc', 'bach'].
        :param grad_dir::
          directory where raw Graduation data is located; defaults to default download dir name.
        :param rm_disk::
          removes downloaded Graduation data from disk, after cleaning.
        '''
        df = CLEANERS[self.subject](graduation_dir=grad_dir, deg_level=degree_level)
        if rm_disk:
            shutil.rmtree(grad_dir)
        return df
    
    def run(self, degree_level='bach', see_progress=False, merge_with_char=False, rm_disk=False) -> pd.DataFrame:
        '''scrapes and cleans IPEDS Graduation data; returns Pandas Dataframe.
        
        :param degree_level::
         level of graduate; options include ['assc', 'bach'].

        :param see_progress::
        (bool) When True, prints successful download confirmation for each year's data. If False, no messages printed.
        
        :param merge_with_char::
        (bool) When True, scrapes Graduation data and merges with Characteristics data (includes variables like school name and address). Characteristics data will automatically be removed from disk when the Dataframe is returned. (To keep the Characteristics data, create a Characteristics object.)
        
        :param rm_disk::
          removes downloaded Graduation data from disk, after cleaning.
        '''
        self.scrape(see_progress=see_progress)
        df = self.clean(rm_disk=rm_disk, degree_level=degree_level)
        if merge_with_char:
            char_df = Characteristics(year_range=self.year_range).run(see_progress=see_progress, rm_disk=True)
            df = df.merge(char_df, on=['id', 'year'])
        return df
    
    


import os
import re
import warnings
from bs4 import BeautifulSoup
import pandas as pd
import numpy as np
import us 

from genpeds.config import VARIABLE_RENAME

def clean_characteristics(characteristics_dir = 'characteristicsdata') -> pd.DataFrame:
    '''cleans institution characteristics data and returns complete characteristics data

    :characteristics_dir:        directory where raw enrollment data is located
    '''
    warnings.filterwarnings('ignore', category=FutureWarning)
    sorted_files = sorted(os.listdir(characteristics_dir))
    rename_dict = VARIABLE_RENAME['characteristics']
    state_mappings = us.states.mapping('abbr', 'name') # get abbr -> names for states
    state_mappings['DC'] = 'District of Columbia' # add Washington D.C. to state mapping

    master_df = pd.DataFrame()

    dtypes = {
        'unitid' : str, 'instnm' : str, 'addr' : str, 'city' : str, 
        'stabbr' : str, 'zip' : str, 'webaddr' : str, 'longitud' : str, 'latitude' : str
    }
    for file in sorted_files:
        file_path = os.path.join(characteristics_dir, file)
        df = pd.read_csv(file_path, dtype=dtypes, encoding_errors='replace', low_memory=False)
        df = df.rename(str.lower, axis='columns')
        year_num = re.split(r'_|\.', f'{file}')[1]
        
        if int(year_num) > 1998:
            if int(year_num) > 2008:
                filt_col = ['unitid', 'instnm', 'addr', 'city', 'stabbr', 'zip', 'webaddr', 'longitud', 'latitude']
            else:
                filt_col = ['unitid', 'instnm', 'addr', 'city', 'stabbr', 'zip', 'webaddr'] # long/lat NA for before 2008
        else:
            filt_col = ['unitid', 'instnm', 'addr', 'city', 'stabbr', 'zip'] # for years < 1999
        df_filtered = df.loc[:, filt_col]
        
        for col in ['instnm', 'addr', 'city']:
            df_filtered[col] = df_filtered[col].str.title() # TitleCase
        df_filtered['year'] = int(year_num) # year identifier
        df_filtered['unitid'] = df_filtered['unitid'].astype(str).str.strip() # make id into string
        df_filtered['stabbr'] = df_filtered['stabbr'].map(state_mappings) # state abbreviation to state name
        master_df = pd.concat([master_df, df_filtered], ignore_index=True)
    
    master_df = master_df.rename(columns=rename_dict) # rename vars
    return master_df


def clean_admissions(admissions_dir = 'admissionsdata') -> pd.DataFrame:
    '''cleans yearly admissions data and returns complete admissions data
    
    :admissions_dir:        directory where raw admissions data is located
    '''
    warnings.filterwarnings('ignore', category=FutureWarning)
    sorted_files = sorted(os.listdir(admissions_dir)) # unnecessary, but helps with error checking
    rename_dict = VARIABLE_RENAME['admissions']

    master_df = pd.DataFrame()
    
    for file in sorted_files:
        file_path = os.path.join(admissions_dir, file)
        df = pd.read_csv(file_path, dtype=str) # read in df
        df = df.rename(str.lower, axis='columns') # some df's have all uppercase, some have all lowercase
        df.columns = df.columns.str.strip() # some column names have right spaces
        year_num = re.split(r'_|\.', f'{file}')[1]
        
        cols_to_filter = [col for col in rename_dict.keys() if col in df.columns] # cols to filter per year
        df_filtered = df.reindex(columns=cols_to_filter)
        df_filtered = df_filtered.rename(columns=rename_dict) # rename cols

        for col in df_filtered.columns:
            if col == 'id':
                df_filtered[col] = df_filtered[col].astype(str).str.strip() # id str
            else:
                df_filtered[col] = pd.to_numeric(df_filtered[col], errors='coerce')

        if int(year_num) == 2001:
            df_filtered['men_enrolled'] = df_filtered['men_ft_enrolled'] + df_filtered['men_pt_enrolled']
            df_filtered['women_enrolled'] = df_filtered['women_ft_enrolled'] + df_filtered['women_pt_enrolled']
        if 'tot_applied' not in df_filtered.columns:
            df_filtered['tot_applied'] = df_filtered['men_applied'] + df_filtered['women_applied']
            df_filtered['tot_admitted'] = df_filtered['men_admitted'] + df_filtered['women_admitted']
            df_filtered['tot_enrolled'] = df_filtered['men_enrolled'] + df_filtered['women_enrolled']

        for i in ['men', 'women']:
            df_filtered[f'accept_rate_{i}'] = np.where(
            df_filtered[f'{i}_applied'] == 0,
            np.nan,
            (df_filtered[f'{i}_admitted'] / df_filtered[f'{i}_applied'] * 100)
            )
    
            df_filtered[f'yield_rate_{i}'] = np.where(
            df_filtered[f'{i}_admitted'] == 0,
            np.nan,
            (df_filtered[f'{i}_enrolled'] / df_filtered[f'{i}_admitted'] * 100)
            )
        
        df_filtered['year'] = int(year_num) # year identifier
        df_filtered['men_applied_share'] = df_filtered['men_applied'] / df_filtered['tot_applied'] * 100
        df_filtered['men_admitted_share'] = df_filtered['men_admitted'] / df_filtered['tot_admitted'] * 100

        master_df = pd.concat([master_df, df_filtered], ignore_index=True)
    
    # unneeded columns
    admissions_df = master_df.drop(columns=['women_applied', 'women_admitted', 'women_enrolled',
                                            'men_ft_enrolled', 'men_pt_enrolled', 'women_ft_enrolled', 'women_pt_enrolled'],
                                   errors='ignore')

    return admissions_df


def clean_enrollment(enrollment_dir = 'enrollmentdata', student_level = 'undergrad') -> pd.DataFrame:
    '''cleans yearly enrollment data and returns complete student enrollment data

    :enrollment_dir:        directory where raw enrollment data is located
    :student_level:        level of enrollment; options include ['undergrad', 'grad']
    '''
    warnings.filterwarnings('ignore', category=FutureWarning)
    sorted_files = sorted(os.listdir(enrollment_dir)) # unnecessary, but helps with error checking
    rename_dict = VARIABLE_RENAME['enrollment']

    grad_rules = [
        (lambda y: y in [1984,1985], 'line in [11,25,10,24]'),
        (lambda y: (y == 1986) or (y in range(1990,1999)), 'line in [14,28,9,10,23,24]'),
        (lambda y: y in [1987,1988,1989], 'line in [14,28]'),
        (lambda y: y == 1999, 'line in [32,52,16]'),
        (lambda y: (y in range(2000,2002)) or (y in range(2004,2009)), 'line in [11,25,9,23]'),
        (lambda y:  y in range(2002,2004), 'line in [12,26]'),
        (lambda y: y in range(2009,2024), 'line in [11,25]')
    ]
    
    master_df = pd.DataFrame()

    for file in sorted_files:
        file_path = os.path.join(enrollment_dir, file)
        df = pd.read_csv(file_path, dtype=str) # read in df
        df = df.rename(str.lower, axis='columns') # some df's have all uppercase, some have all lowercase
        year_num = re.split(r'_|\.', f'{file}')[1]

        if all(col in df.columns for col in ['efrace10', 'eftotlm']):
            cols_to_filter = [col for col in rename_dict.keys() if 
                              (col in df.columns) and ('efrace' not in col)] # annoying thing with duplicate cols in 
                                                                                # some years
        else:
            cols_to_filter = [col for col in rename_dict.keys() if col in df.columns]
            

        df_filtered = df.reindex(columns=cols_to_filter)
        df_filtered = df_filtered.rename(columns=rename_dict) # rename cols

        for col in df_filtered.columns:
            if col == 'id':
                df_filtered[col] = df_filtered[col].astype(str).str.strip() # id identifier
            else:
                df_filtered[col] = pd.to_numeric(df_filtered[col], errors='coerce')

        if student_level == 'undergrad':
            if int(year_num) < 1986:
                student_query = 'line == 1 or line == 15' # captures total full-time and total part-time undergrads, respectively
            else:
                student_query = 'line == 8 or line == 22' 
        elif student_level == 'grad':
            for cond,frmt in grad_rules:
                if cond(int(year_num)):
                    student_query = frmt # captures full-time and part-time graduate and first-professional students
                    break
            else:
                raise ValueError(f'No formatted rule for year {year_num}')
        else:
            raise ValueError("student_level must be 'undergrad' or 'grad' ")

        students = df_filtered.query(student_query) # filter data to total students
        
        if 'wtmen' not in students.columns:
            cols_to_sum = ['totmen', 'totwomen']
        else:
            cols_to_sum = ['totmen', 'totwomen', 'wtmen', 'wtwomen','bkmen', 'bkwomen','hspmen', 'hspwomen','asnmen', 'asnwomen']
        
        students_by_inst = students.groupby('id')[cols_to_sum].sum() # sum full-time and part-time students by school
        students_by_inst = students_by_inst.eval('totmen_share = totmen / (totmen + totwomen) * 100').reset_index() # male student share
        students_by_inst['year'] = int(year_num) # get year marker for each set
        students_by_inst['studentlevel'] = student_level # get student level identifier

        if 'wtmen' in students_by_inst.columns:
            for attr in ['wt', 'bk', 'hsp', 'asn']:
                eval_str = f'tot{attr}_share = ({attr}men + {attr}women) / (totmen + totwomen) * 100' # race share breakdowns
                students_by_inst = students_by_inst.eval(eval_str)

        master_df = pd.concat([master_df, students_by_inst], ignore_index=True)

    return master_df


def clean_completion(completion_dir = 'completiondata', level = 'bach') -> pd.DataFrame:
    '''cleans yearly completion data and returns complete completions data

    :completion_dir:        directory where raw completion data is located
    :level:                 level of degree, options include ['assc', 'bach', 'mast', 'doct']
    '''
    warnings.filterwarnings('ignore', category=FutureWarning)
    sorted_files = sorted(os.listdir(completion_dir)) # unnecessary, but helps with error checking
    rename_dict = VARIABLE_RENAME['completion']

    deglevel_rules = [
        (lambda l,y: l == 'assc', 'awlevel == 3'),
        (lambda l,y: l == 'bach', 'awlevel == 5'),
        (lambda l,y: l == 'mast', 'awlevel == 7'),
        (lambda l,y: (l == 'doct') and (y < 2010), 'awlevel == 9'),
        (lambda l,y: (l == 'doct') and (y >= 2010), 'awlevel >= 17 and awlevel <= 19')
    ]
    
    master_df = pd.DataFrame()

    for file in sorted_files:
        file_path = os.path.join(completion_dir, file)
        df = pd.read_csv(file_path, dtype=str) # read in df
        df = df.rename(str.lower, axis='columns') # some df's have all uppercase, some have all lowercase
        year_num = re.split(r'_|\.', f'{file}')[1]
        
        if all(col in df.columns for col in ['crace10', 'ctotalm']):
            cols_to_filter = [col for col in rename_dict.keys() if 
                              (col in df.columns) and ('crace' not in col)] # annoying thing with duplicate cols in 
                                                                                # some years
        else:
            cols_to_filter = [col for col in rename_dict.keys() if col in df.columns]

        df_filtered = df.reindex(columns=cols_to_filter)
        df_filtered = df_filtered.rename(columns=rename_dict)
        for col in df_filtered:
            if col not in ['id', 'cip']:
                df_filtered[col] = pd.to_numeric(df_filtered[col], errors='coerce')
        
        for cond,frmt in deglevel_rules:
            if cond(level, int(year_num)):
                level_query = frmt
                break
        else:
            raise ValueError("level must be 'assc', 'bach', 'mast' or 'doct'") 
        
        completions = df_filtered.query(level_query)
        race_cols = [col for col in completions.columns if 'men' in col] # race columns to group
        completions = completions.groupby(['id', 'cip'])[race_cols].sum().reset_index()

        completions = completions.eval('totmen_share = totmen / (totmen + totwomen) * 100') # maleshare within each major
        if 'wtmen' in completions.columns:
            for attr in ['wt', 'bk', 'hsp', 'asn']:
                eval_str = f'tot{attr}_share = ({attr}men + {attr}women) / (totmen + totwomen) * 100' # race share breakdowns
                completions = completions.eval(eval_str)
        
        completions['deglevel'] = level # adds level identifier
        completions['year'] = int(year_num) # adds year identifier
        completions['cip'] = completions['cip'].astype(str).str.strip()
        
        master_df = pd.concat([master_df, completions], ignore_index=True)

    return master_df


def clean_cip_html(file_path):
    '''returns dict of CIP subject code:label pairs for a given year's CIP dictionary html.
    
    :file_path: string path to CIP data dictionary html
    '''
    with open(file_path) as filehandle:
        soup = BeautifulSoup(filehandle, 'html.parser')
        rows = soup.find_all('tr', attrs={'bgcolor': ['White', 'Silver']})
        dat_rows = rows[1:]
        label_dict = {}
        for row in dat_rows:
            cells = row.find_all('td')
            if len(cells) > 1:
                label = cells[0].text.strip()
                val = cells[1].text.strip()
                if label == 'Totals':
                    break # end of relevant table
                else:
                    label_dict[val] = label
        return label_dict


def clean_cip(cip_codes_dir = 'cipdata') -> pd.DataFrame:
    '''cleans yearly CIP data and returns full dataframe

    :cip_codes_dir: directory where raw CIP data is located
    '''
    warnings.filterwarnings('ignore', category=FutureWarning)
    warnings.filterwarnings('ignore', category=UserWarning)
    sorted_files = sorted(os.listdir(cip_codes_dir))
    master_df = pd.DataFrame()

    for file in sorted_files:
        file_path = os.path.join(cip_codes_dir, file)
        year_num = re.split(r'_|\.', f'{file}')[1]
        ext = re.split(r'_|\.', f'{file}')[2]
        if ext == 'html':
            html_dict = clean_cip_html(file_path=file_path)
            df = pd.DataFrame({'cip_description' : html_dict.values(),
                               'cip' : html_dict.keys()}, dtype=str)
        else:
            df = pd.read_excel(file_path, sheet_name='Frequencies', dtype=str)
            df = df.query('varname == "CIPCODE" or varname == "Cipcode"').loc[:, ['codevalue', 'valuelabel']]
            df = df.rename(columns={'codevalue' : 'cip', 'valuelabel' : 'cip_description'})
        df['year'] = int(year_num) # year identifier
        
        master_df = pd.concat([master_df, df], ignore_index=True)
        master_df['cip'] = master_df['cip'].astype(str).str.strip() # strip spaces
        master_df['cip_description'] = master_df['cip_description'].str.title().replace(r'^(\d+)\s-\s', '', regex=True)

    return master_df


def clean_graduation(graduation_dir = 'graduationdata', deg_level='bach') -> pd.DataFrame:
    '''cleans yearly graduation data and returns complete graduation data

    :graduation_dir:        directory where raw completion data is located
    :deg_level:        degree level; options include ['assc', 'bach']
    '''
    warnings.filterwarnings('ignore', category=FutureWarning)
    sorted_files = sorted(os.listdir(graduation_dir)) # unnecessary, but helps with error checking
    rename_dict = VARIABLE_RENAME['graduation']
    master_df = pd.DataFrame()

    for file in sorted_files:
        file_path = os.path.join(graduation_dir, file)
        df = pd.read_csv(file_path, dtype=str) # read in df
        df = df.rename(str.lower, axis='columns') # some df's have all uppercase, some have all lowercase
        year_num = re.split(r'_|\.', f'{file}')[1]

        if all(col in df.columns for col in ['grrace10', 'grtotlm']):
            cols_to_filter = [col for col in rename_dict.keys() if 
                              (col in df.columns) and ('grrace' not in col)] # annoying thing with duplicate cols in 
                                                                                # some years
        else:
            cols_to_filter = [col for col in rename_dict.keys() if col in df.columns]
        df_filtered = df.reindex(columns=cols_to_filter)
        df_filtered = df_filtered.rename(columns=rename_dict)
        
        for col in df_filtered.columns:
            if col != 'id':
                df_filtered[col] = pd.to_numeric(df_filtered[col], errors='coerce') # convert cols to float
        
        if deg_level == 'bach':
            deg_query = '(8 <= grtype <= 9) and (12 <= chrtstat <= 13) and section == 2'
            denom = 8
            num = 9
        elif deg_level == 'assc':
            deg_query = '(29 <= grtype <= 30) and (12 <= chrtstat <= 13) and section == 4'
            denom = 29
            num = 30
        else:
            raise  ValueError("deg_level must be 'assc', 'bach', 'mast' or 'doct'")

        grads = df_filtered.query(deg_query)
        pivoted_grads = grads.pivot(index='id', columns='grtype', values=['totmen', 'totwomen', # get cohort and grads
                                                                                        'wtmen', 'wtwomen',
                                                                                        'bkmen', 'bkwomen',
                                                                                        'hspmen', 'hspwomen',
                                                                                        'asnmen', 'asnwomen'])
        for i in ['tot', 'wt', 'bk', 'hsp', 'asn']:
            pivoted_grads[f'gradrate_{i}men'] = (pivoted_grads[f'{i}men'][num] /    
                                                     pivoted_grads[f'{i}men'][denom] * 100)
            
            pivoted_grads[f'gradrate_{i}women'] = (pivoted_grads[f'{i}women'][num] / 
                                                     pivoted_grads[f'{i}women'][denom] * 100) # grad rates
        pivoted_grads = pivoted_grads.reset_index()

        rnm_columns = pivoted_grads.columns.droplevel(1).tolist() # renaming columns
        for i in range(len(rnm_columns)):
            if rnm_columns[i] == rnm_columns[i - 1]:
                rnm_columns[i] = f'{rnm_columns[i]}_graduated'
        pivoted_grads.columns = rnm_columns
        
        pivoted_grads['year'] = int(year_num) # get year identifiers
        pivoted_grads['deglevel'] = deg_level

        master_df = pd.concat([master_df, pivoted_grads], ignore_index=True)
    
    return master_df
        

CLEANERS = {
    'characteristics' : clean_characteristics,
    'admissions' : clean_admissions,
    'enrollment' : clean_enrollment,
    'completion' : clean_completion,
    'cip' : clean_cip,
    'graduation' : clean_graduation
}



import concurrent.futures
import time
import random
import requests
import zipfile
import os
import warnings
import re
from genpeds.config import DATASETS

def get_file_endpoint(subject, year):
    '''returns endpoint for a given subject in a given year.
    
    :param year: year for file; available years vary by subject.
    :param subject: subject.
    '''
    format_rules = DATASETS[subject]['format_rules'] # rules for each subject-year combination
    endpoint_template = DATASETS[subject]['file_template'] # endpoint template

    lag0, lag1, lead1 = year-1900, year-1901, year-1899 # needed for format rules
    for cond,frmt in format_rules:
        if cond(year):
            yr_format = frmt.format(year=year, 
                                        lag0=str(lag0), 
                                        lag1 = str(lag1), 
                                        lead1 = str(lead1))
            endpoint = endpoint_template.format(yr_format) # formatted endpoint
            break
    else:
            warnings.warn(f'No formatted rule for year {year}')
            return None
    
    return endpoint

def download_a_file(subject, year):
    '''downloads an IPEDS subject-year data file.

    :param year: year for file; available years vary by subject.
    :param subject: subject.
    '''
    relevant_dir = DATASETS[subject]['dir'] # directory subject name
    relevant_prefix = DATASETS[subject]['file_prefix'] # file subject prefix

    endpoint = get_file_endpoint(subject, year) # get endpoint for a subject-year combination

    try:
        r = requests.get(endpoint)  # try request
    except requests.HTTPError as er:
        return f"Year {year}: Error - {str(er)}"
        
    if '404 - File or directory not found' in r.text:
        return f"Year {year}: 404 - File not found"
    
    zipped_file = os.path.join(relevant_dir, f'{relevant_prefix}_{year}.zip')
    open(zipped_file, 'wb').write(r.content)
    
    with zipfile.ZipFile(zipped_file, 'r') as zfile:
        if subject == 'cip':
            file_to_extract = endpoint.split('/')[-1].replace('_Dict.zip', '').lower()
            for ext in ['.html', '.xls', '.xlsx']:  # diff file formats, try each one
                try:
                    zfile.extract(file_to_extract + ext, relevant_dir)
                    # remove & remove zipped file
                    old_name_file = os.path.join(relevant_dir, f'{file_to_extract}{ext}')
                    new_name_file = os.path.join(relevant_dir, f'cipcodes_{year}{ext}')
                    os.rename(old_name_file, new_name_file)
                    os.remove(zipped_file)
                    break
                except KeyError:
                    continue
        else:
            file_to_extract = endpoint.split('/')[-1].replace('.zip', '').lower() + '.csv'
            zfile.extract(file_to_extract, relevant_dir)
            # remove & remove zipped file
            old_name_file = os.path.join(relevant_dir, file_to_extract)
            new_name_file = os.path.join(relevant_dir, f'{relevant_prefix}_{year}.csv')
            os.rename(old_name_file, new_name_file)
            os.remove(zipped_file)

    return(f'IPEDS {subject.title()} ({year}) successfully downloaded and extracted')


def scrape_ipeds_data(subject='characteristics', year_range = None, see_progress = True):
    '''downloads NCES IPEDS data on specified years for a defined subject.
    
    :param subject: string identifying which subject data to download. The subjects available are:
     ['characteristics', 'admissions', 'enrollment', 'completion', 'cip', 'graduation']
    
    :param year_range: tuple of year integers (indicates a range), iterable of year integers (indicates group of individual years), or single year to pull data from. Data for 'characteristics', 'enrollment' and 'completion' are available for years 1984-2023, while 'graduation' is available for years 2000-2023. Defaults to all available years for a subject.

    :param see_progress: boolean that, when true, prints completion statement for extraction of each year. If false, no messages printed.
    
    ## available data

    - :characteristics: institutional characteristics, like a school's name, address. Certain variables, like a school's longitude and latitude are only available in later years. Available for years 1984-2023.

    - :admissions: Admissions data, like number of applications and acceptances by gender. Available for years 2001-2023.

    - :enrollment: fall enrollment by gender and institutional level (e.g., 4-year undergraduate program), with most years including enrollment by race and gender. Available for years 1984-2023.

    - :completion: completion of degrees by gender, level of degree and subject field (e.g., Bachelor's in Economics), with most years including completion by race and gender. Available for years 1984-2023.

    - :cip: CIP, or Classification of Instructional Programs, are key-value pairs for subject study fields. CIP's vary by year, and are relevant to identify subject field in completion data. Available for years 1984-2023.

    - :graduation: number of cohorts and graduates by gender, institutional level and graduation measure (e.g., students earning a bachelor's degree within 6 years of entering). Available for years 2000-2023.
    '''
    subject = subject.lower()
    relevant_dir = DATASETS[subject]['dir']
    relevant_prefix = DATASETS[subject]['file_prefix'] # file subject prefix
    # Determine the years to download
    if not year_range:
        if subject == 'graduation':
            start, end = 2000, 2023
            iter_range = range(start, end + 1)  # default for graduation data
        elif subject == 'admissions':
            start, end = 2001, 2023
            iter_range = range(start, end + 1)  # default for admissions data
        else:
            start, end = 1984, 2023
            iter_range = range(start, end + 1)  # default for all other data
    else:
        if isinstance(year_range, tuple):
            start, end = year_range
            iter_range = range(start, end + 1)  # tuple follows regular range
        elif isinstance(year_range, list):
            iter_range = year_range  # list remains list
        elif isinstance(year_range, int):
            start = year_range
            iter_range = [start]  # integer becomes one-element list
        else:
            raise ValueError('Please enter a tuple range, list of integers, or a single integer')
    
    if os.path.isdir(relevant_dir): # so we don't need to redownload if it isn't necessary
        iter_range2 = []
        stripped_list = [re.sub(r'\.csv|\.html|\.xlsx|\.xlsx','',ff) 
                             for ff in sorted(os.listdir(relevant_dir))]
        for yr in iter_range:
            if f'{relevant_prefix}_{yr}' not in stripped_list:
                iter_range2.append(yr)
        iter_range = iter_range2
    else:
        os.makedirs(relevant_dir, exist_ok=True) # create subject directory, where unzipped files will be stored

    # multithread to speed up the process
    with concurrent.futures.ThreadPoolExecutor(max_workers=5) as exec:
        future_to_year = {exec.submit(download_a_file, subject, year): year for year in iter_range}
        for future in concurrent.futures.as_completed(future_to_year):
            yr = future_to_year[future]
            try:
                result = future.result()
                if see_progress:
                    print(result) # if you want to see the progress
                time.sleep(random.uniform(0.1, 0.3)) # you're welcome NCES :)
            except Exception as exc:
                print(f"Year {yr} generated an exception: {exc}")


if __name__=='__main__':
    scrape_ipeds_data(subject='enrollment',
                      year_range=[2000,2002,2003,2010,2020,2023,2017],
                      see_progress=True)



'''
MAIN DATASETS AS OF FIRST RELEASE
- INCLUDES [CHARACTERISTICS, ADMISSIONS, ENROLLMENT, COMPLETION, GRADUATION]
- DICTIONARY INCLUDES DESCRIPTION AND ENDPOINT FORMATTING RULES
'''
DATASETS = {

    'characteristics' : {
        'description' : '''Institutional characteristics for each school in a given year, including information like a school's name and address; available for years 1984-2023.''',
        'years_available' : (1984,2023),
        'dir' : 'characteristicsdata',
        'file_prefix' : 'characteristics',
        'file_template' : 'https://nces.ed.gov/ipeds/datacenter/data/{}.zip', 
        'format_rules' : [
            (lambda y : y <= 1985, 'IC{year}'),
            (lambda y : (1985 < y <= 1989) or (1991 < y <= 1994), 'IC{year}_A'),
            (lambda y : y == 1990,  'IC90HD'), 
            (lambda y : y == 1991,  'IC1991_hdr'),
            (lambda y : 1994 < y < 1997, 'ic{lag0}{lead1}_A'), 
            (lambda y : y == 1997, 'ic9798_HDR'), 
            (lambda y : y == 1998, 'IC98hdac'),
            (lambda y : y == 1999, 'IC99_HD'),
            (lambda y : y == 2000 or y == 2001, 'FA{year}HD'),
            (lambda y : y >= 2002, 'HD{year}')
        ]
        
    },

    'admissions' : {
        'description' : '''Admissions data for each school in a given year (measured in the Fall), including information like ACT/SAT scores and acceptance rates by gender; available for years 2001-2023.''',
        'years_available' : (2001,2023),
        'dir' : 'admissionsdata',
        'file_prefix' : 'admissions',
        'file_template' : 'https://nces.ed.gov/ipeds/datacenter/data/{}.zip',
        'format_rules' : [
            (lambda y: 2001 <= y <= 2013, 'IC{year}'),
            (lambda y: 2014 <= y <= 2023, 'ADM{year}'), 
        ]
    },

    'enrollment' : {
        'description' : '''Enrollment data for each school in a given year (measured in the Fall), including information like the male enrollment share and enrollment by gender and race; available at the undergraduate or graduate level; available for years 1984-2023.''',
        'years_available' : (1984,2023),
        'dir' : 'enrollmentdata',
        'file_prefix' : 'enrollment',
        'file_template' : 'https://nces.ed.gov/ipeds/datacenter/data/{}.zip',
        'format_rules' : [
            (lambda y: y <= 1985, 'EF{year}'),
            (lambda y: (1985 < y < 1990) or (1991 < y < 1994), 'EF{year}_A'),
            (lambda y: y == 1990, 'EF90_A'),
            (lambda y: y == 1991, 'ef1991_A'),
            (lambda y: y == 1994,  'EF{year}_ANR'),
            (lambda y: 1994 < y < 2000, 'EF{lag0}_ANR'),
            (lambda y: y >= 2000, 'EF{year}A')
        ]
    },

    'completion' : {
        'description' : '''Completion data for each school in a given year by detailed subject field, including information like the male completion share within a subject (e.g., Economics); available at the Associate, Bachelor's, Master's or Doctoral level; available for years 1984-2023.''',
        'years_available' : (1984,2023),
        'dir' : 'completiondata',
        'file_prefix' : 'completion',
        'file_template' : 'https://nces.ed.gov/ipeds/datacenter/data/{}.zip',
        'format_rules' : [
            (lambda y: y < 1990 or (1991 < y < 1995), 'C{year}_CIP'),
            (lambda y: y == 1990, 'C8990CIP'),
            (lambda y: y == 1991, 'c1991_cip'),
            (lambda y: 1995 <= y < 2000, 'C{lag1}{lag0}_A'),
            (lambda y: y >= 2000, 'C{year}_A')
        ]
    },

    'graduation' : {
        'description' : '''Graduation data for each school in a given year, including information like the graduation rate by gender and race; available at the Associate or Bachelor's level; available for years 2000-2023.''',
        'years_available' : (2000,2023),
        'dir' : 'graduationdata',
        'file_prefix' : 'graduation',
        'file_template' : 'https://nces.ed.gov/ipeds/datacenter/data/{}.zip',
        'format_rules' : [
            (lambda y: y in range(2000,2024), 'GR{year}')
        ]
    },

    'cip' : {
        'description' : '''Provides subject field codes for subjects in a given year; should be used with Completion data; available for years 1984-2023.''',
        'years_available' : (1984,2023),
        'dir' : 'cipdata',
        'file_prefix' : 'cip',
        'file_template' : 'https://nces.ed.gov/ipeds/datacenter/data/{}_Dict.zip',
        'format_rules' : [
            (lambda y: (y < 1990) or (1991 < y < 1995), 'C{year}_CIP'),
            (lambda y: y == 1990, 'C8990CIP'),
            (lambda y: y == 1991, 'c1991_cip'),
            (lambda y: 1995 <= y < 2000, 'C{lag1}{lag0}_A'),
            (lambda y: y >= 2000, 'C{year}_A')
        ]
    }

}


'''
VARIABLES AND VARIABLE DESCRIPTIONS INCLUDED IN EACH SUBJECT DATASET
'''
VARIABLE_DICT = {

    'characteristics' : {
       'id' : 'Institutional unique identifier ID; though UNITIDs can vary over time due to instutions splitting, this variable is generally consistent for an institution over time.', 
       'year' : 'Survey year.',
       'name' : 'Name of institution location.', 
       'address' : 'Address of institution location.', 
       'city' : 'City of institution location.', 
       'state' : 'State of institution location.', 
       'zipcode' : 'Zipcode of institution location.', 
       'webaddress' : 'Admissions website URL of institution location.',
       'longitude' : 'Longitude coordinate of institution location.', 
       'latitude' : 'Latitude coordinate of institution location.', 
    },

    'admissions' : {
       'id' : 'Institutional unique identifier ID; though UNITIDs can vary over time due to instutions splitting, this variable is generally consistent for an institution over time.', 
       'year' : 'Survey year.',
       'tot_applied' : 'Total number of first-time, degree/certificate-seeking applicants in a given year.',
       'tot_admitted' : 'Total number of first-time, degree/certificate-seeking applicants who were admitted in a given year.', 
       'tot_enrolled' : 'Total number of admitteees who ultimately enrolled (includes part-time and full-time enrollemnt).', 
       'men_applied' : 'Number of first-time, degree/certificate-seeking men who applied to an institution in a given year.', 
       'men_admitted' : 'Number of first-time, degree/certificate-seeking men who were admitted to an institution in a given year.', 
       'men_enrolled' : 'Number of admitted men who ultimately enrolled (includes part-time and full-time enrollemnt).', 
       'accept_rate_men' : 'Acceptance rate of male applicants.', 
       'accept_rate_women' : 'Acceptance rate of female applicants.', 
       'yield_rate_men': 'Yield rate (share of admitted students who ultimately enrolled) of male admittees.',
       'yield_rate_women' : 'Yield rate (share of admitted students who ultimately enrolled) of female admittees.',
       'men_applied_share' : 'Male share of total number of first-time, degree/certificate-seeking applicants in a given year.',
       'men_admitted_share' : 'Male share of total number of first-time, degree/certificate-seeking admittees in a given year.', 
       'share_submit_sat' : 'Share of admitted students who submitted an SAT score.',
       'share_submit_act' : 'Share of admitted students who submitted an ACT score.', 
       'sat_rw_25' : 'SAT reading/writing score (25th percentile).',
       'sat_rw_50' : 'SAT reading/writing score (50th percentile).', 
       'sat_rw_75' : 'SAT reading/writing score (75th percentile).', 
       'sat_math_25' : 'SAT mathematics score (25th percentile).',
       'sat_math_50' : 'SAT mathematics score (50th percentile).',
       'sat_math_75' : 'SAT mathematics score (75th percentile).', 
       'act_comp_25' : 'ACT composite score (25th percentile).', 
       'act_comp_50' : 'ACT composite score (50th percentile).',
       'act_comp_75' : 'ACT composite score (75th percentile).', 
       'act_eng_25' : 'ACT English score (25th percentile).', 
       'act_eng_50' : 'ACT English score (50th percentile).', 
       'act_eng_75' : 'ACT English score (75th percentile).',
       'act_math_25' : 'ACT mathematics score (25th percentile).', 
       'act_math_50' : 'ACT mathematics score (50th percentile).', 
       'act_math_75' : 'ACT mathematics score (75th percentile).'
    },

    'enrollment' : {
       'id' : 'Institutional unique identifier ID; though UNITIDs can vary over time due to instutions splitting, this variable is generally consistent for an institution over time.', 
       'year' : 'Survey year.',
       'studentlevel' : "Level of student enrollment; either 'undergrad' (Undergraduate) or 'grad' (Graduate).",
       'totmen' : 'Total number of men enrolled in the Fall; this includes both part-time and full-time enrollment.', 
       'totwomen' : 'Total number of women enrolled in the Fall; this includes both part-time and full-time enrollment.', 
       'wtmen' : 'Total number of non-Hispanic White men enrolled in the Fall; this includes both part-time and full-time enrollment.', 
       'wtwomen' : 'Total number of non-Hispanic White women enrolled in the Fall; this includes both part-time and full-time enrollment.', 
       'bkmen' : 'Total number of non-Hispanic Black men enrolled in the Fall; this includes both part-time and full-time enrollment.',
       'bkwomen' : 'Total number of non-Hispanic Black women enrolled in the Fall; this includes both part-time and full-time enrollment.',
       'hspmen' : 'Total number of Hispanic men (of any race) enrolled in the Fall; this includes both part-time and full-time enrollment.', 
       'hspwomen' : 'Total number of Hispanic women (of any race) enrolled in the Fall; this includes both part-time and full-time enrollment.', 
       'asnmen' : 'Total number of non-Hispanic Asian men enrolled in the Fall; this includes both part-time and full-time enrollment.', 
       'asnwomen' : 'Total number of non-Hispanic Asian women enrolled in the Fall; this includes both part-time and full-time enrollment.', 
       'totmen_share' : "Male share of total enrollment; 'total' here is defined as the sum of male and female enrollment; this includes both part-time and full-time enrollment.",  
       'totwt_share' : "Non-Hispanic White share of total enrollment; 'total' here is defined as the sum of male and female enrollment; this includes both part-time and full-time enrollment.",   
       'totbk_share' : "Non-Hispanic Black share of total enrollment; 'total' here is defined as the sum of male and female enrollment; this includes both part-time and full-time enrollment.", 
       'tothsp_share' : "Hispanic share of total enrollment; 'total' here is defined as the sum of male and female enrollment; this includes both part-time and full-time enrollment.",
       'totasn_share' : "Non-Hispanic Asian share of total enrollment; 'total' here is defined as the sum of male and female enrollment; this includes both part-time and full-time enrollment."
    },

    'completion' : {
        'id' : 'Institutional unique identifier ID; though UNITIDs can vary over time due to instutions splitting, this variable is generally consistent for an institution over time.', 
        'year' : 'Survey year. Note that the cohort being tracked is the cohort that started either six years back (for bachelor) or three years back (for associate).',
        'cip' : 'CIP (Classification of Instructional Programs) code for a subject in a given year.', 
        'totmen' : 'Total number of male completers within a CIP subject.', 
        'totwomen' : 'Total number of female completers within a CIP subject.', 
        'wtmen' : 'Total number of non-Hispanic White men completers within a CIP subject.; this includes both part-time and full-time enrollment.', 
        'wtwomen' : 'Total number of non-Hispanic White women completers within a CIP subject.; this includes both part-time and full-time enrollment.', 
        'bkmen' : 'Total number of non-Hispanic Black men completers within a CIP subject.; this includes both part-time and full-time enrollment.',
        'bkwomen' : 'Total number of non-Hispanic Black women completers within a CIP subject.; this includes both part-time and full-time enrollment.',
        'hspmen' : 'Total number of Hispanic men (of any race) completers within a CIP subject.; this includes both part-time and full-time enrollment.', 
        'hspwomen' : 'Total number of Hispanic women (of any race) completers within a CIP subject.; this includes both part-time and full-time enrollment.', 
        'asnmen' : 'Total number of non-Hispanic Asian men completers within a CIP subject.; this includes both part-time and full-time enrollment.', 
        'asnwomen' : 'Total number of non-Hispanic Asian women completers within a CIP subject.; this includes both part-time and full-time enrollment.', 
        'totwt_share' : "Non-Hispanic White share of total (male and female) completers within a CIP subject.",   
        'totbk_share' : "Non-Hispanic Black share of total (male and female) completers within a CIP subject.", 
        'tothsp_share' : "Hispanic share of total (male and female) completers within a CIP subject.",
        'totasn_share' : "Non-Hispanic Asian share of total (male and female) completers within a CIP subject.",
        'totmen_share' : 'Male share of total (male and female) completers within a CIP subject.', 
        'deglevel' : "Degree level of completers; options include 'assc' (Associate's), 'bach' (Bachelor's), 'mast' (Master's), and 'doct' (Doctoral)."
    },

    'cip' : {
        'year' : 'Survey year. Note that the cohort being tracked is the cohort that started either six years back (for bachelor) or three years back (for associate).',
        'cip' : 'CIP (Classification of Instructional Programs) identifier code.',
        'cip_description' : 'CIP subject label (e.g., Geology).'

    },

    'graduation' : {
       'id' : 'Institutional unique identifier ID; though UNITIDs can vary over time due to instutions splitting, this variable is generally consistent for an institution over time.', 
       'year' : 'Survey year. Note that the cohort being tracked is the cohort that started either six years back (for bachelor) or three years back (for associate).',
       'deglevel' : "Level of student degree; either 'bach' (Bachelor's) or 'assc' (Associates).",
       'totmen' : 'Total number of men within an adjusted cohort.', 
       'totmen_graduated' : "Total number of men within an adjusted cohort that graduated within 150 percent of the normal time considered to graduate; for bachelor's degrees, this is six years, and for associate's degrees, this is three years.", 
       'totwomen' : 'Total number of women within an adjusted cohort.', 
       'totwomen_graduated' : "Total number of women within an adjusted cohort that graduated within 150 percent of the normal time considered to graduate; for bachelor's degrees, this is six years, and for associate's degrees, this is three years.",
       'wtmen' : 'Total number of non-Hispanic White men within an adjusted cohort.', 
       'wtmen_graduated' : "Total number of non-Hispanic White men within an adjusted cohort that graduated within 150 percent of the normal time considered to graduate; for bachelor's degrees, this is six years, and for associate's degrees, this is three years.", 
       'wtwomen' : 'Total number of non-Hispanic White women within an adjusted cohort.', 
       'wtwomen_graduated' : "Total number of  non-Hispanic White women within an adjusted cohort that graduated within 150 percent of the normal time considered to graduate; for bachelor's degrees, this is six years, and for associate's degrees, this is three years.", 
       'bkmen' : 'Total number of non-Hispanic Black men within an adjusted cohort.',
       'bkmen_graduated' : "Total number of non-Hispanic Black men within an adjusted cohort that graduated within 150 percent of the normal time considered to graduate; for bachelor's degrees, this is six years, and for associate's degrees, this is three years.", 
       'bkwomen' : 'Total number of non-Hispanic Black women within an adjusted cohort.', 
       'bkwomen_graduated' : "Total number of non-Hispanic Black women within an adjusted cohort that graduated within 150 percent of the normal time considered to graduate; for bachelor's degrees, this is six years, and for associate's degrees, this is three years.", 
       'hspmen' : 'Total number of Hispanic men within an adjusted cohort.',
       'hspmen_graduated' : "Total number of Hispanic men within an adjusted cohort that graduated within 150 percent of the normal time considered to graduate; for bachelor's degrees, this is six years, and for associate's degrees, this is three years.", 
       'hspwomen' : 'Total number of Hispanic women within an adjusted cohort.', 
       'hspwomen_graduated' : "Total number of Hispanic women within an adjusted cohort that graduated within 150 percent of the normal time considered to graduate; for bachelor's degrees, this is six years, and for associate's degrees, this is three years.", 
       'asnmen' : 'Total number of non-Hispanic Asian men within an adjusted cohort.',
       'asnmen_graduated' : "Total number of non-Hispanic Asian men within an adjusted cohort that graduated within 150 percent of the normal time considered to graduate; for bachelor's degrees, this is six years, and for associate's degrees, this is three years.", 
       'asnwomen' : 'Total number of non-Hispanic Asian women within an adjusted cohort.', 
       'asnwomen_graduated' : "Total number of non-Hispanic Asian women within an adjusted cohort that graduated within 150 percent of the normal time considered to graduate; for bachelor's degrees, this is six years, and for associate's degrees, this is three years.",  
       'gradrate_totmen' : 'Graduation rate for men (within 150 percent of normal time taken to graduate).',
       'gradrate_totwomen' : 'Graduation rate for women (within 150 percent of normal time taken to graduate).', 
       'gradrate_wtmen' : 'Graduation rate for non-Hispanic White men (within 150 percent of normal time taken to graduate).', 
       'gradrate_wtwomen' : 'Graduation rate for non-Hispanic White women (within 150 percent of normal time taken to graduate).',
       'gradrate_bkmen' : 'Graduation rate for non-Hispanic Black men (within 150 percent of normal time taken to graduate).', 
       'gradrate_bkwomen' : 'Graduation rate for non-Hispanic Black women (within 150 percent of normal time taken to graduate).', 
       'gradrate_hspmen' : 'Graduation rate for Hispanic men (within 150 percent of normal time taken to graduate).',
       'gradrate_hspwomen' : 'Graduation rate for Hispanic women (within 150 percent of normal time taken to graduate).', 
       'gradrate_asnmen' : 'Graduation rate for  non-Hispanic Asianmen (within 150 percent of normal time taken to graduate).', 
       'gradrate_asnwomen' : 'Graduation rate for non-Hispanic Asian women (within 150 percent of normal time taken to graduate).' 
    }
    
}

VARIABLE_RENAME = {
    'characteristics' : {
        'unitid' : 'id', 'instnm' : 'name',
          'addr' : 'address', 'city' : 'city', 'stabbr' : 'state', 'zip' : 'zipcode', 
          'webaddr' : 'webaddress', 
          'longitud' : 'longitude', 'latitude' : 'latitude'
    },

    'admissions' : {
        'unitid' : 'id', 
        'applcnm' : 'men_applied', 'applcnw' : 'women_applied', 
        'admssnm' : 'men_admitted', 'admssnw' : 'women_admitted',
        'satpct' : 'share_submit_sat', 'actpct' : 'share_submit_act',                                      
        'satvr25' : 'sat_rw_25', 'satvr75' : 'sat_rw_75', 
        'satmt25' : 'sat_math_25', 'satmt75' : 'sat_math_75', 
        'actcm25' : 'act_comp_25', 'actcm75' : 'act_comp_75',
        'acten25' : 'act_eng_25', 'acten75' : 'act_eng_75', 
        'actmt25' : 'act_math_25', 'actmt75' : 'act_math_75',
        'enrlftm' : 'men_ft_enrolled', 'enrlftw' : 'women_ft_enrolled',
        'enrlptm' : 'men_pt_enrolled', 'enrlptw' : 'women_pt_enrolled',
        'enrlm' : 'men_enrolled', 'enrlw' : 'women_enrolled',
        'applcn' : 'tot_applied', 'admssn' : 'tot_admitted', 'enrlt' : 'tot_enrolled',
        'acten50' : 'act_eng_50', 'actmt50' : 'act_math_50', 'actcm50' : 'act_comp_50',
        'satvr50' : 'sat_rw_50', 'satmt50' : 'sat_math_50'
    },

    'enrollment' : {
        'unitid' : 'id', 'line' : 'line', 'efrace15' : 'totmen', 'efrace16' : 'totwomen',
        'eftotlm' : 'totmen', 'eftotlw' : 'totwomen', 'efwhitm' : 'wtmen', 'efwhitw' : 'wtwomen',
        'efbkaam' : 'bkmen', 'efbkaaw' : 'bkwomen', 'efhispm' : 'hspmen', 'efhispw' : 'hspwomen', 
        'efasiam' : 'asnmen', 'efasiaw' : 'asnwomen', 'eftotlw' : 'totwomen',
        'efrace11' : 'wtmen', 'efrace12' : 'wtwomen', 'efrace03' : 'bkmen', 'efrace04' : 'bkwomen', 
        'efrace09' : 'hspmen', 'efrace10' : 'hspwomen', 'efrace07' : 'asnmen', 'efrace08' : 'asnwomen'
    },

    'completion' : {
        'unitid' : 'id', 'cipcode' : 'cip', 'awlevel' : 'awlevel',
        'crace15' : 'totmen', 'crace16' : 'totwomen',
        'ctotalm' : 'totmen', 'ctotalw' : 'totwomen', 
        'cwhitm' : 'wtmen', 'cwhitw' : 'wtwomen', 'cbkaam' : 'bkmen', 'cbkaaw' : 'bkwomen', 
        'chispm' : 'hspmen', 'chispw' : 'hspwomen', 'casiam' : 'asnmen', 'casiaw' : 'asnwomen',
        'crace11' : 'wtmen', 'crace12' : 'wtwomen', 'crace03' : 'bkmen', 'crace04' : 'bkwomen', 
        'crace09' : 'hspmen', 'crace10' : 'hspwomen', 'crace07' : 'asnmen', 'crace08' : 'asnwomen'
    },

    'cip' : {
        # nothing needed for now.
    },

    'graduation' : {
        'grtotlm' : 'totmen', 'grtotlw' : 'totwomen', 
        'grwhitm' : 'wtmen', 'grwhitw' : 'wtwomen',
        'grbkaam' : 'bkmen', 'grbkaaw' : 'bkwomen', 
        'grhispm' : 'hspmen', 'grhispw' : 'hspwomen', 
        'grasiam' : 'asnmen', 'grasiaw' : 'asnwomen', 
        'grrace15' : 'totmen', 'grrace16' : 'totwomen',
        'grrace11' : 'wtmen', 'grrace12' : 'wtwomen', 
        'grrace03' : 'bkmen', 'grrace04' : 'bkwomen', 
        'grrace09' : 'hspmen', 'grrace10' : 'hspwomen', 
        'grrace07' : 'asnmen', 'grrace08' : 'asnwomen',
        'chrtstat' : 'chrtstat', 'section' : 'section', 
        'cohort' : 'cohort', 'unitid' : 'id', 'grtype' : 'grtype'
    }
}


from genpeds.cleaners import CLEANERS
from genpeds import scrape_ipeds_data
from genpeds.config import VARIABLE_DICT
import pandas as pd
import os
import glob
import shutil
import pytest

def download_data_for_test():
    '''downloads data for test, assuming not already downloaded'''
    downloads = [
        ('characteristics', [2000,2010,2020]),
        ('admissions', (2001,2005)),
        ('enrollment', (1985,1990)),
        ('completion', [2002,2022,2023]),
        ('cip', [2002,2022,2023]),
        ('graduation', (2003,2008))
    ]
    for sbjct, yrs in downloads:
        if os.path.exists(f'{sbjct}data'):
            if isinstance(yrs, tuple):
                start,end = yrs
                iter_range = range(start, end + 1)
            else:
                iter_range = yrs
            for yr in iter_range:
                pattern = os.path.join(f'{sbjct}data', f'{sbjct}_{yr}.*')
                matches = glob.glob(pattern)
                if not matches:
                    scrape_ipeds_data(subject=sbjct, year_range=yr, see_progress=False)
                else:
                    continue
        else:
            scrape_ipeds_data(subject=sbjct, year_range=yrs, see_progress=False)

# some subjects take no arguments (other than directory path)
# test these together
@pytest.mark.parametrize('subject', [
    'characteristics',
    'admissions',
    'cip'
])
def test_basic_cleaners(subject):
    '''test cleaner functions that take no arguments'''
    download_data_for_test() # scrape data if needed

    subject_cleaner = CLEANERS[subject] # subject-specific cleaner function
    subject_var_dict = VARIABLE_DICT[subject].keys() # expected variables returned

    try:
        df = subject_cleaner()
        assert isinstance(df, pd.DataFrame), f'{subject} dataframe not returned.' # check that Pandas DataFrame is returned

        for col in df.columns:
            assert col in subject_var_dict # check if attributes are in expected attributes
        shutil.rmtree(f'{subject}data')
    finally:
        pass

# test enrollment, with variable arguments
@pytest.mark.parametrize('lev', [
    'undergrad',
    'grad'
])
def test_enrollment_cleaner(lev):
    '''test enrollment cleaning function'''
    download_data_for_test() # scrape data if needed

    subject_cleaner = CLEANERS['enrollment'] # enrollment cleaner function
    subject_var_dict = VARIABLE_DICT['enrollment'].keys() # expected variables returned

    try:
        df = subject_cleaner(student_level=lev) # tries 'undergrad' and 'grad'
        assert isinstance(df, pd.DataFrame), f'Enrollment dataframe not returned.' # check that Pandas DataFrame is returned

        for col in df.columns:
            assert col in subject_var_dict # check if attributes are in expected attributes
        shutil.rmtree(f'enrollmentdata')
    finally:
        pass

# test completion
@pytest.mark.parametrize('lev', [
    'assc', 
    'bach', 
    'mast', 
    'doct'
])
def test_completion_cleaner(lev):
    '''test completion cleaning function'''
    download_data_for_test() # scrape data if needed

    subject_cleaner = CLEANERS['completion'] # completion cleaner function
    subject_var_dict = VARIABLE_DICT['completion'].keys() # expected variables returned

    try:
        df = subject_cleaner(level=lev) # tries 'assc', 'bach', 'mast', 'doct'
        assert isinstance(df, pd.DataFrame), f'Completion dataframe not returned.' # check that Pandas DataFrame is returned

        for col in df.columns:
            assert col in subject_var_dict # check if attributes are in expected attributes
        shutil.rmtree(f'completiondata')
    finally:
        pass

# test graduation
@pytest.mark.parametrize('lev', [
    'assc', 
    'bach'
])
def test_graduation_cleaner(lev):
    '''test graduation cleaning function'''
    download_data_for_test() # scrape data if needed

    subject_cleaner = CLEANERS['graduation'] # graduation cleaner function
    subject_var_dict = VARIABLE_DICT['graduation'].keys() # expected variables returned

    try:
        df = subject_cleaner(deg_level=lev) # tries 'undergrad' and 'grad'
        assert isinstance(df, pd.DataFrame), f'Graduation dataframe not returned.' # check that Pandas DataFrame is returned

        for col in df.columns:
            assert col in subject_var_dict # check if attributes are in expected attributes
        shutil.rmtree(f'graduationdata')
    finally:
        pass


'''Work with gendered NCES IPEDS data: from admissions to graduation'''

__version__ = '1.0.2'

from genpeds.core import Characteristics, Admissions, Enrollment, Completion, Cip, Graduation
from genpeds.downloader import scrape_ipeds_data

__all__ = [
    'Characteristics',
    'Admissions',
    'Enrollment', 
    'Completion',
    'Cip',
    'Graduation',
    'scrape_ipeds_data'
]
